#!/bin/bash
set -e

if [ $# -lt 2 ]; then
    echo "Usage ./test install_folder hysop_folder [cache_dir]"
    exit 1
fi

if [ $# -gt 3 ]; then
    echo "Usage ./test install_folder hysop_folder [cache_dir]"
    exit 1
fi

if [ ! -d "$1" ]; then
    echo "Folder $1 does not exist."
    exit 1
fi

if [ ! -d "$2" ]; then
    echo "Folder $2 does not exist."
    exit 1
fi

INSTALL_DIR=$1
HYSOP_DIR=$2
if [ $# -eq 3 ]; then
    CACHE_DIR=$3
    HAS_CACHE_DIR=true
else
    HAS_CACHE_DIR=false
fi

if [ "$HAS_CACHE_DIR" = true ]; then
    if [ -d "$CACHE_DIR" ]; then
        echo "Cache directory '$CACHE_DIR' was found."
        mkdir -p /root/.cache
        cp -r $CACHE_DIR/* /root/.cache
    else
        echo "Cache directory '$CACHE_DIR' was not found."
        mkdir -p $CACHE_DIR
    fi
fi

export PYTHONPATH="$INSTALL_DIR/lib/python2.7/site-packages:$INSTALL_DIR:$PYTHONPATH"
export MPLBACKEND='cairo'
export HYSOP_VERBOSE=0
export HYSOP_DEBUG=0
export HYSOP_PROFILE=0
export HYSOP_KERNEL_DEBUG=0

# OpenMPI specific variables
export OMPI_ALLOW_RUN_AS_ROOT=1
export OMPI_ALLOW_RUN_AS_ROOT_CONFIRM=1
export OMPI_MCA_rmaps_base_oversubscribe=1

echo "Trying to load hysop module:"
python -c 'import hysop; print hysop'
echo "module import successful !"
echo

DO_TESTS=${DO_TESTS:-true}
DO_EXAMPLES=${DO_EXAMPLES:-true}
DO_LONG_TESTS=${DO_LONG_TESTS:-false}

COMMON_TEST_OPTIONS=''
TEST_DIR="$HYSOP_DIR"
COMMON_EXAMPLE_OPTIONS='-VNC -cp default -maxit 2'
EXAMPLE_DIR="$HYSOP_DIR/../examples"

hysop_test() {
     test=$1
     echo 
     echo "TESTING $1"
     echo "========$(printf '=%.0s' `seq ${#1}`)"
     python2 "${TEST_DIR}/${1}" ${@:2} ${COMMON_TEST_OPTIONS} 
     echo
}
example_test() {
     test=$1
     echo 
     echo "EXAMPLE $1"
     echo "========$(printf '=%.0s' `seq ${#1}`)"
     python2 "${EXAMPLE_DIR}/${1}" ${@:2} ${COMMON_EXAMPLE_OPTIONS}
     echo
}

if [ "$DO_TESTS" = true ]; then
    hysop_test "core/arrays/tests/test_array.py"
    hysop_test "core/graph/tests/test_graph.py"
    hysop_test "fields/tests/test_fields.py"
    hysop_test "numerics/tests/test_fft.py"
    hysop_test "operator/tests/test_analytic.py"
    hysop_test "operator/tests/test_transpose.py"
    hysop_test "operator/tests/test_fd_derivative.py"
    hysop_test "operator/tests/test_absorption.py"
    hysop_test "operator/tests/test_penalization.py"
    hysop_test "operator/tests/test_velocity_correction.py"
    hysop_test "operator/tests/test_restriction_filter.py"
    hysop_test "operator/tests/test_directional_advection.py"
    hysop_test "operator/tests/test_directional_diffusion.py"
    hysop_test "operator/tests/test_directional_stretching.py"
    hysop_test "operator/tests/test_custom_symbolic.py"
    hysop_test "operator/tests/test_spectral_derivative.py"
    hysop_test "operator/tests/test_spectral_curl.py"
    hysop_test "operator/tests/test_diffusion.py"
    hysop_test "operator/tests/test_poisson.py"
    hysop_test "operator/tests/test_solenoidal_projection.py"
    hysop_test "operator/tests/test_poisson_curl.py"
    $HYSOP_DIR/fields/tests/test_cartesian.sh
fi

if [ "$DO_LONG_TESTS" = true ]; then
    hysop_test "backend/device/codegen/kernels/tests/test_directional_advection.py"
    hysop_test "backend/device/codegen/kernels/tests/test_directional_remesh.py"
fi

if [ "$DO_EXAMPLES" = true ]; then
    example_test "analytic/analytic.py"
    example_test "scalar_diffusion/scalar_diffusion.py"
    example_test "scalar_advection/scalar_advection.py"
    example_test "scalar_advection/levelset.py"
    example_test "multiresolution/scalar_advection.py"
    example_test "shear_layer/shear_layer.py"
    example_test "taylor_green/taylor_green.py" '-impl python'
    example_test "taylor_green/taylor_green.py" '-impl opencl'
    example_test "bubble/periodic_bubble.py"
    example_test "bubble/periodic_bubble_levelset.py"
    example_test "bubble/periodic_bubble_levelset_penalization.py"
    example_test "bubble/periodic_jet_levelset.py"
    example_test "particles_above_salt/particles_above_salt_periodic.py"
    example_test "particles_above_salt/particles_above_salt_symmetrized.py"
fi

if [ "$HAS_CACHE_DIR" = true ]; then
    cp -r /root/.cache/* $CACHE_DIR/
    find $CACHE_DIR -name '*.lock' -delete
fi

exit 0
