from hysop.deps import sm, np, contextlib
from hysop.tools.types import check_instance, first_not_None
from hysop.backend.device.opencl.opencl_types           import OpenClTypeGen

from hysop.backend.device.codegen.base.opencl_codegen   import OpenClCodeGenerator
from hysop.backend.device.codegen.base.function_codegen import OpenClFunctionCodeGenerator
from hysop.backend.device.codegen.base.variables        import CodegenVariable, CodegenVectorClBuiltin
from hysop.backend.device.codegen.base.utils            import WriteOnceDict, ArgDict
from hysop.backend.device.codegen.base.statistics       import WorkStatistics

from hysop.backend.device.codegen.functions.polynomial import PolynomialFunction
from hysop.backend.device.codegen.functions.custom_atomics import CustomAtomicFunction

from hysop.constants import BoundaryCondition
from hysop.numerics.remesh.remesh import RemeshKernel

from hysop.numerics.stencil.stencil_generator import StencilGenerator

class DirectionalRemeshFunction(OpenClFunctionCodeGenerator):

    def __init__(self, typegen, work_dim, itype, ftype,
            nparticles, nscalars,
            sboundary, remesh_kernel,
            use_atomics, remesh_criteria_eps,
            use_short_circuit=None,
            known_args=None,
            debug_mode=False):

        check_instance(sboundary,tuple,values=BoundaryCondition)
        check_instance(remesh_kernel, RemeshKernel)
        assert remesh_kernel.n % 2 == 0 or remesh_kernel.n == 1
        assert remesh_kernel.n > 0

        use_short_circuit = first_not_None(use_short_circuit, typegen.use_short_circuit_ops)

        is_periodic     = (sboundary[0]==BoundaryCondition.PERIODIC) and \
                          (sboundary[1]==BoundaryCondition.PERIODIC)

        ivtype = typegen.vtype(itype, nparticles)
        fvtype = typegen.vtype(ftype, nparticles)

        reqs = self.build_requirements(typegen, work_dim, itype, ftype,
                ivtype, fvtype, nparticles, nscalars, sboundary, remesh_kernel,
                use_atomics, remesh_criteria_eps)

        (args,basename) = self.build_prototype(reqs, typegen, work_dim, itype, ftype,
                ivtype, fvtype, nparticles, nscalars, sboundary, remesh_kernel,
                use_atomics, remesh_criteria_eps, is_periodic)

        super(DirectionalRemeshFunction,self).__init__(basename=basename,
                output=None, typegen=typegen, inline=True,
                args=args, known_args=known_args)

        self.update_requirements(reqs)

        self.work_dim = work_dim
        self.itype  = itype
        self.ftype  = ftype
        self.ivtype = ivtype
        self.fvtype = fvtype

        self.nparticles = nparticles
        self.nscalars = nscalars
        self.sboundary = sboundary
        self.kernel = remesh_kernel
        self.use_atomics = use_atomics
        self.remesh_criteria_eps = remesh_criteria_eps
        self.is_periodic = is_periodic
        self.debug_mode = debug_mode

        self.use_short_circuit = use_short_circuit

        self.gencode()

    @staticmethod
    def _printv(ncomponents):
        assert ncomponents in [1,2,4,8,16]
        if ncomponents==1:
            return ''
        else:
            return 'v{}'.format(ncomponents)


    def build_prototype(self, reqs, typegen, work_dim, itype, ftype,
                ivtype, fvtype, nparticles, nscalars, sboundary, remesh_kernel,
                use_atomics, remesh_criteria_eps, is_periodic):

        args = ArgDict()

        atomic  = 'atomic_'  if use_atomics else ''
        criteria = '{}eps__'.format(remesh_criteria_eps) if (remesh_criteria_eps is not None) else 'full'

        basename =  '__{}remesh_{}d__lambda_{}_{}__{}__{}p__{}s__{}'.format(
                atomic, work_dim,
                remesh_kernel.n, remesh_kernel.r, ftype,
                nparticles, nscalars,
                criteria)

        args['p'] = CodegenVectorClBuiltin(name='p', btype=ftype, dim=nparticles, typegen=typegen,
                add_impl_const=True, nl=True)

        scalars = []
        for i in xrange(nscalars):
            Si = CodegenVectorClBuiltin(name='s{}'.format(i), btype=ftype, dim=nparticles, typegen=typegen,
                        add_impl_const=True, nl=(i==nscalars-1))
            scalars.append(Si)
            args[Si.name] = Si

        args['dx'] = CodegenVariable(name='dx', ctype=ftype, typegen=typegen,
                    add_impl_const=True)
        args['inv_dx'] = CodegenVariable(name='inv_dx', ctype=ftype, typegen=typegen,
                    add_impl_const=True, nl=True)

        if is_periodic:
            args['grid_size'] = CodegenVariable(name='grid_size', ctype=itype, typegen=typegen,
                        add_impl_const=True)
        args['line_offset'] = CodegenVariable(name='line_offset', ctype=itype, typegen=typegen,
                    add_impl_const=True)
        args['cache_ghosts'] = CodegenVariable(name='cache_ghosts', ctype=itype, typegen=typegen,
                    add_impl_const=True)
        args['active'] = CodegenVariable(name='active', ctype='bool', typegen=typegen,
                    add_impl_const=True, nl=True)


        cached_scalars = []
        for i in xrange(nscalars):
            Si = CodegenVariable(name='S{}'.format(i), ctype=ftype, typegen=typegen,
                        ptr_restrict=True, ptr=True, storage=self._local,
                        ptr_const=False, add_impl_const=True, nl=True)
            cached_scalars.append(Si)
            args[Si.name] = Si

        self.scalars = scalars
        self.cached_scalars = cached_scalars

        return (args,basename)


    def build_requirements(self, typegen, work_dim, itype, ftype,
                ivtype, fvtype, nparticles, nscalars, sboundary, remesh_kernel,
                use_atomics, remesh_criteria_eps):

        reqs = WriteOnceDict()

        P = []
        kernel = remesh_kernel
        if kernel.poly_splitted:
            self.poly_splitted = True
            name='lambda_{}_{}__{}'.format(kernel.n, kernel.r, '{}_{}')
            for i, _ in enumerate(kernel.Pt_l):
                ilname = name.format(i,'left')
                irname = name.format(i,'right')
                Pil = PolynomialFunction(typegen, ftype, nparticles,
                        kernel.Cl[:,i], ilname, use_fma=True, var='y')
                Pir = PolynomialFunction(typegen, ftype, nparticles,
                        kernel.Cr[:,i], irname, use_fma=True, var='y')
                P.append((Pil,Pir))
                reqs[ilname] = Pil
                reqs[irname] = Pir
        else:
            self.poly_splitted = False
            name='lambda_{}_{}__{}'.format(kernel.n, kernel.r, '{}')
            for i, _ in enumerate(kernel.Pt):
                iname = name.format(i)
                Pi = PolynomialFunction(typegen, ftype, nparticles,
                        kernel.C[:,i], iname, use_fma=True, var='y')
                P.append((Pi,))
                reqs[iname] = Pi
        self.polynomials = P

        if use_atomics:
            reqs['atomic_add'] = CustomAtomicFunction(typegen, ftype, '{} + {}', self._local, 'add')

        return reqs

    def gencode(self):
        s = self
        dim = s.work_dim
        tg = self.typegen

        itype = s.itype
        ftype = s.ftype
        ivtype = s.ivtype
        fvtype = s.fvtype

        work_dim = s.work_dim
        itype = s.itype
        ftype = s.ftype
        ivtype = s.ivtype
        fvtype = s.fvtype

        nparticles  = s.nparticles
        nscalars    = s.nscalars
        sboundary   = s.sboundary
        use_atomics = s.use_atomics
        use_short_circuit = s.use_short_circuit
        remesh_criteria_eps = s.remesh_criteria_eps
        debug_mode = s.debug_mode

        is_periodic = s.is_periodic

        scalars = s.scalars
        cached_scalars = s.cached_scalars

        pos = s.vars['p']

        dx = s.vars['dx']
        inv_dx = s.vars['inv_dx']

        active = s.vars['active']
        line_offset = s.vars['line_offset']
        cache_ghosts = s.vars['cache_ghosts']

        poly_splitted = s.poly_splitted

        printf_no_vector=True
        if printf_no_vector:
            vnf = '[{}]'.format(', '.join('%2.2f' for _ in xrange(nparticles)))
            vni = '[{}]'.format(', '.join('%i' for _ in xrange(nparticles)))
            expand_printf_vector = lambda x: ','.join(x[i] for i in xrange(nparticles))
        else:
            vnf = '[%2.2{}f]'.format(self._printv(nparticles))
            vni = '[%{}i]'.format(self._printv(nparticles))
            expand_printf_vector = lambda x: str(x)
        epv = expand_printf_vector

        @contextlib.contextmanager
        def if_thread_active():
            with s._if_('{}'.format(active)):
                yield

        if is_periodic:
            grid_size = s.vars['grid_size']

        dtype = tg.np_dtype(ftype)

        P   = CodegenVariable(name='P', ctype=itype, typegen=tg,
                const=True, value=1+(self.kernel.n/2))
        eps = CodegenVariable(name='eps', ctype=ftype, typegen=tg,
                const=True, value=np.finfo(dtype).eps)

        rp   = CodegenVectorClBuiltin(name='rp',    btype=ftype, dim=nparticles, typegen=tg)
        y    = CodegenVectorClBuiltin(name='y',    btype=ftype, dim=nparticles, typegen=tg)
        ind  = CodegenVectorClBuiltin(name='ind',  btype=itype, dim=nparticles, typegen=tg)
        find = CodegenVectorClBuiltin(name='find', btype=ftype, dim=nparticles, typegen=tg)
        vone = CodegenVectorClBuiltin(name='one', btype=ftype, dim=nparticles, typegen=tg, value=(1,)*nparticles)

        tst0 = CodegenVectorClBuiltin(name='tst0',    btype=ftype, dim=nparticles, typegen=tg)
        tst1 = CodegenVectorClBuiltin(name='tst1',    btype=ftype, dim=nparticles, typegen=tg)

        if poly_splitted:
            wl = CodegenVectorClBuiltin(name='Wl', btype=ftype, dim=nparticles, typegen=tg)
            wr = CodegenVectorClBuiltin(name='Wr', btype=ftype, dim=nparticles, typegen=tg)
            w = CodegenVectorClBuiltin(name='W', btype=ftype, dim=nparticles, typegen=tg)
            weights = (wl,wr,w)
        else:
            w = CodegenVectorClBuiltin(name='W', btype=ftype, dim=nparticles, typegen=tg)
            weights = (w,)

        with s._function_():
            P.declare(s)
            if (remesh_criteria_eps is not None):
                eps.declare(s)
            s.jumpline()

            #s.decl_aligned_vars(tst0, tst1)
            s.decl_aligned_vars(rp, find, ind, y)
            vone.declare(s, const=True)
            s.decl_vars(*weights)
            s.jumpline()

            with if_thread_active():
                s.append('{} = {}*{};'.format(rp, pos, inv_dx))
                s.append('{} = fract({}, &{});'.format(y, rp, find))
                s.append('{} = convert_{}_rtn({});'.format(ind, ivtype, find))
                #s.append('{} = floor({rp});'.format(tst0, rp=rp))
                #s.append('{} = {rp} - floor({rp});'.format(tst1, rp=rp))
                if debug_mode:
                    with self._ordered_wi_execution_(barrier=False):
                        code = 'printf("%lu p={vnf}, rp={vnf}, floor(rp)={vnf}, rp-floor(rp)={vnf}, ind={vni}, y={vnf}, s0={vnf}, a=%f, b=%f, c=%f.\\n", {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {});'.format(
                                'get_local_id(0)', epv(pos), epv(rp), epv(tst0), epv(tst1), epv(ind), epv(y), epv(scalars[0]), 'floor(4.5f)', 'floor((float2)(1.5f,2.54f)).x', 'floor((float2)(1.5f,2.54f)).y',
                                vnf=vnf, vni=vni)
                        s.append(code)
                y.affect(s, init='{}-{}'.format(vone,y))
                ind.affect(s, init='{} - {} - {}'.format(ind, P, line_offset))

            for i, Pi in enumerate(self.polynomials):
                s.jumpline()
                with if_thread_active():
                    s.append('{} += 1;'.format(ind))
                    with s._align_() as al:
                        for wj,Pij,yj in zip(weights, Pi, (y,'1-{}'.format(y))):
                            wj.affect(al, align=True, init=Pij(y=yj))
                        if poly_splitted:
                            if nparticles>1:
                                w.affect(al, align=True, init='select({}, {}, ({}<{}))'.format(
                                   wr, wl, y, tg.dump(0.5)))
                            else:
                                w.affect(al, align=True, init='convert_{fvtype}({}<{})*{} + convert_{fvtype}({}>={})*{}'.format(
                                    y,tg.dump(0.5), wl, y, tg.dump(0.5), wr, fvtype=fvtype))

                    for iscal, (cached_scalar, scalar) in enumerate(zip(cached_scalars, scalars)):
                        if nparticles>1:
                            comment='Remeshing scalar {}, {} particles at a time.'.format(iscal, nparticles)
                            s.jumpline()
                            s.comment(comment)
                        with s._block_():
                            for ipart in xrange(nparticles):
                                cache_idx='{}+{}'.format(cache_ghosts, ind[ipart])
                                val = '{}*{}'.format(w[ipart],scalar[ipart])
                                if (remesh_criteria_eps is None):
                                    criteria = '{}!={}'.format(scalar[ipart], tg.dump(0.0))
                                else:
                                    criteria = 'fabs({}) > {}*eps'.format(scalar[ipart], remesh_criteria_eps)

                                if use_atomics:
                                    atomic_add = s.reqs['atomic_add']
                                    atom_add = atomic_add(p='{}+{}'.format(cached_scalar,cache_idx), val=val)
                                    if use_short_circuit:
                                        code = '({}) && ({},true);'.format(criteria, atom_add)
                                    else:
                                        code = 'if({}) {{ {}; }}'.format(criteria, atom_add)
                                else:
                                    if use_short_circuit:
                                        code = '({}) && (({} += {}),true);'.format(criteria, cached_scalar[cache_idx], val)
                                    else:
                                        code = 'if ({}) {{ {} += {}; }}'.format(criteria, cached_scalar[cache_idx], val)
                                s.append(code)
                    if debug_mode:
                        val='{}*{}'.format(w,scalars[0])
                        cache_idx='{}+{}'.format(cache_ghosts, ind[0])

                        if poly_splitted:
                            printf = 'printf("BATCH {}: %lu wrote {vnf} at idx={vni} with Wl={vnf}, Wr={vnf}, W={vnf}, cond={vni}.\\n",{},{},{},{},{},{},{});'.format(
                                    i, 'get_local_id(0)', val, ind, wl, wr, w, 'y<0.5f', vnf=vnf, vni=vni)
                        else:
                            printf = 'printf("BATCH {}: %lu wrote {vnf} at idx {vni} with W={vnf}.\\n",{},{},{},{});'.format(
                                    i, 'get_local_id(0)', val, ind,  w, vni=vni, vnf=vnf)
                        s.append(printf)
                if not use_atomics:
                    s.barrier(_local=True)
            if use_atomics:
                s.barrier(_local=True)

    # def per_work_statistics(self):
        # dim     = self.dim
        # ftype   = self.ftype
        # storage = self.storage

        # stats = self.reqs['apply_stencil'].per_work_statistics()

        # if 'alpha' in self.known_args:
            # alpha = self.known_args['alpha']
        # else:
            # alpha = 0.5
        # stats.ops_per_type[ftype] += int(alpha!=0)*dim*3
        # stats.ops_per_type[ftype] += int(alpha!=1)*dim*2 + 2

        # retur
if __name__ == '__main__':
    from hysop.backend.device.codegen.base.test import _test_typegen
    from hysop.numerics.remesh.remesh import RemeshKernel

    tg = _test_typegen('float')

    kernel = RemeshKernel(2, 2, split_polys=True)

    work_dim = 3
    nparticles = 2
    nscalars = 2

    use_atomics = True
    remesh_criteria_eps = None

    drf = DirectionalRemeshFunction(tg, work_dim, 'int', tg.fbtype,
            nparticles, nscalars,
            (BoundaryCondition.PERIODIC,BoundaryCondition.PERIODIC),
            kernel, use_atomics, remesh_criteria_eps)
    drf.edit()
