"""Description of an iterative method (iteration, stopping criteria, ...)"""

from hysop import Simulation, Problem
from hysop.parameters.scalar_parameter import ScalarParameter
from hysop import dprint, vprint
from hysop.tools.decorators  import debug
from hysop.core.graph.graph import ready
from hysop.constants import HYSOP_REAL, HYSOP_INTEGER
from hysop.simulation import eps
from hysop.tools.numpywrappers import npw
import numpy as np

class PseudoSimulation(Simulation):
    """Pseudo time-iterations for iterative method"""

    def __init__(self, stop_criteria, tolerance=1e-8, **kwds):
        """
        Parameters
        ----------
        stop_criteria : TensorParameter
            Iterative loop test for stopping is stop_criteria<tolerance
        tolerance : float (optional)
            Tolerance for stopping criteria (default to 1e-8)

        Notes
        -----
        This object implement an iterative method in a pseudo-time
        inside a 'real' time interval. It can be used as sub-iterations
        for a fixed-point operator.
        It implements a loop defined as follows:

        .. code::

            while(t<end and iteration<max_iter and max(stop_criteria)>tolerance)

        """
        super(PseudoSimulation, self).__init__(**kwds)
        self.stop_criteria, self.tolerance = stop_criteria, tolerance

    def advance(self, dbg=None, plot_freq=10):
        """Proceed to next time.

        * Advance time and iteration number.
        * Compute the new timestep
        * check if simulation is over.
        """
        super(PseudoSimulation, self).advance(dbg=dbg, plot_freq=plot_freq)
        if np.max(self.stop_criteria.value) <= self.tolerance:
            self.is_over = True
            return

    def print_state(self, verbose=None):
        """Print current iteration parameters
        """
        if isinstance(self.stop_criteria, ScalarParameter):
            crit = "{0:6.5g}".format(self.stop_criteria.value)
        else:
            crit = np.array2string(self.stop_criteria.value,
                                   formatter={'float_kind':lambda x: "%g" % x})
        msg = "=== PseudoSimulation : {0:3d}, criteria = {1} =="
        if verbose:
            print msg.format(self.current_iteration, crit)
        else:
            vprint(msg.format(self.current_iteration, crit))


class IterativeMethod(Problem):
    """Overriding a Problem to enfoce a PseudoSimulation for iterative method loop.

    The PseudoSimulation is created on each 'apply'. There is no meaning for the
    underlying pseudo-time.

    Notes
    -----

    Sub-timestepping is not a usual use-case for this Problem. One should
    override Problem class in a proper way. Here only a pseudo-timestep is
    used together with a maximal number of iteration to compute a pseudo-final time.
    """
    def __init__(self, stop_criteria, tolerance=1e-8, state_print=100, max_iter=10000,
                 dt0=None, dt=None, **kwargs):
        super(IterativeMethod, self).__init__(**kwargs)
        self.stop_criteria, self.tolerance = stop_criteria, tolerance

        # create a pseudo-time step parameter if not given.
        if (dt is None):
            dt = ScalarParameter(name='pseudo-dt', dtype=HYSOP_REAL, min_value=eps,
                                 initial_value=eps, quiet=True)
        else:
            dt.value = eps
        self.dt0, self.dt = dt0, dt
        self.state_print = state_print
        self.max_iter = max_iter

        # Store the iteration number of each call to 'apply' in a parameter.
        self.it_num = ScalarParameter(name='iterations', dtype=HYSOP_INTEGER,
                                      initial_value=0, quiet=True)

        # stop_criteria reset value
        self._stop_criteria_reset = npw.ones(self.stop_criteria.shape,
                                             dtype=self.stop_criteria.dtype)
        self._stop_criteria_reset[...] = 1e10

    @debug
    @ready
    def apply(self, simulation, **kwds):
        vprint('=== Entering iterative method...')
        self.stop_criteria.value = self._stop_criteria_reset

        # Initialize the pseudo-simulation for iterative loop
        # end time is computed from pseudo-time step and a maximal number of iterations.
        loop = PseudoSimulation(stop_criteria=self.stop_criteria,
                                tolerance=self.tolerance,
                                start=simulation.t(), end=simulation.t()+self.max_iter*self.dt0,
                                max_iter=self.max_iter,
                                dt0=self.dt0, dt=self.dt,
                                t=ScalarParameter(name='dummy-t',
                                                  dtype=HYSOP_REAL,
                                                  quiet=True))
        # Usual initialize-loop-finalize sequence :
        loop.initialize()
        while not loop.is_over:
            if loop.current_iteration % self.state_print ==0:
                loop.print_state()
            super(IterativeMethod, self).apply(simulation=loop, **kwds)
            loop.advance(dbg=kwds['dbg'])
        vprint("=== Leaving iterative method ({0} iterations, criteria = {1} )".format(
            loop.current_iteration,
            np.array2string(loop.stop_criteria.value, formatter={'float_kind':lambda x: "%g" % x})))
        self.it_num.value = loop.current_iteration
        loop.finalize()

    def get_preserved_input_fields(self):
        return set()
